import logging

import sqlparse

logger = logging.getLogger(__name__)


class SQLRunnerException(Exception):
    pass


def clean_sql_code(code):
    output = ""
    for line in code.split("\n"):
        stripped_line = line.strip()
        if stripped_line == "\\timing":
            continue
        if stripped_line.startswith("--"):
            continue
        output += stripped_line + "\n"
    return output


class Block(object):
    def __init__(self):
        self.closed = False
        self.content = ""

    def append_line(self, line):
        if self.closed:
            raise SQLRunnerException("Block closed !")
        self.content += line

    def close(self):
        if self.closed:
            raise SQLRunnerException("Block closed !")
        self.closed = True

    def is_closed(self):
        return self.closed

    def run(self, connection):
        statements = sqlparse.parse(self.content)

        if "".join((unicode(stmt) for stmt in statements)) != self.content:
            raise SQLRunnerException("sqlparse failed to properly split input")

        rows = 0
        with connection.cursor() as cursor:
            for statement in statements:
                if clean_sql_code(str(statement)).strip() in ("", ";"):
                    # Sometimes sqlparse keeps the empty lines here,
                    # this could negatively affect libpq
                    continue
                logger.debug("Running one statement... <<%s>>", str(statement))
                cursor.execute(str(statement).replace("\\timing\n", ""))
                logger.debug("Affected %s rows", cursor.rowcount)
                rows += cursor.rowcount
        return rows


class SimpleBlock(Block):
    def run(self, connection):
        with connection.cursor() as cursor:
            cursor.execute(str(self.content))


class MetaBlock(Block):
    def __init__(self, command):
        super(MetaBlock, self).__init__()
        self.command = command
        if command != "do-until-0":
            raise SQLRunnerException("Unexpected command {}".format(command))

    def run(self, db):
        total_rows = 0
        # Simply call super().run in a loop...
        delta = 0
        batch_delta = -1
        while batch_delta != 0:
            batch_delta = 0
            logger.debug("Running one block in a loop")
            delta = super(MetaBlock, self).run(db)
            if delta > 0:
                total_rows += delta
                batch_delta = delta
            logger.debug("Batch delta done : %s", batch_delta)
        return total_rows


class Script(object):
    def __init__(self, file_handle):
        is_manual = '/manual/' in file_handle.name
        if not is_manual:
            self.block_list = [SimpleBlock()]
        else:
            self.block_list = [Block()]
        for line in file_handle:
            if line.startswith("--meta-psql:") and is_manual:
                self.block_list[-1].close()
                command = line.split(":")[1].strip()
                if command == 'done':
                    # create a new basic block
                    self.block_list.append(Block())
                else:
                    # create a new meta block
                    self.block_list.append(MetaBlock(command))
            else:
                self.block_list[-1].append_line(line)
        self.block_list[-1].close()

    def run(self, db):
        for block in self.block_list:
            block.run(db)



# -*- coding: utf-8 -*-
import io
import logging
import os.path

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connection

from django_north.management import migrations
from django_north.management.runner import Script

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Migrate the DB to the target version."

    def handle(self, *args, **options):
        if getattr(settings, 'NORTH_MANAGE_DB', False) is not True:
            logger.info('migrate command disabled')
            return

        self.verbosity = options.get('verbosity')

        self.migrate()

    def migrate(self):
        # build miration plan
        migration_plan = migrations.build_migration_plan()

        if migration_plan is None:
            # schema not inited
            self.init_schema()
            # reload migration_plan
            migration_plan = migrations.build_migration_plan()

        # play migrations
        recorder = migrations.MigrationRecorder(connection)
        for plan in migration_plan['plans']:
            version = plan['version']
            if self.verbosity >= 1:
                self.stdout.write(self.style.MIGRATE_LABEL(version))
            for mig, applied, path in plan['plan']:
                title = mig
                if '/manual/' in path:
                    title += ' (manual)'
                if applied:
                    if self.verbosity >= 1:
                        self.stdout.write("  {} already applied".format(title))
                else:
                    if self.verbosity >= 1:
                        self.stdout.write("  Applying {}...".format(title))
                    self.run_script(path)
                    recorder.record_applied(version, mig)

    def init_schema(self):
        init_version = migrations.get_version_for_init()

        # load additional files
        additional_files = getattr(
            settings, 'NORTH_ADDITIONAL_SCHEMA_FILES', [])
        for file_name in additional_files:
            file_path = os.path.join(
                settings.NORTH_MIGRATIONS_ROOT, 'schemas', file_name)
            if self.verbosity >= 1:
                self.stdout.write(
                    self.style.MIGRATE_LABEL("Load {}".format(file_name)))
            self.run_script(file_path)

        # load schema
        if self.verbosity >= 1:
            self.stdout.write(self.style.MIGRATE_LABEL("Load schema"))
            self.stdout.write("  Applying {}...".format(init_version))
        schema_path = os.path.join(
            settings.NORTH_MIGRATIONS_ROOT,
            'schemas',
            getattr(settings, 'NORTH_SCHEMA_TPL',
                    migrations.schema_default_tpl)
            .format(init_version))
        self.run_script(schema_path)

        # load fixtures
        if self.verbosity >= 1:
            self.stdout.write(self.style.MIGRATE_LABEL("Load fixtures"))
            self.stdout.write("  Applying {}...".format(init_version))
        fixtures_path = os.path.join(
            settings.NORTH_MIGRATIONS_ROOT,
            'fixtures',
            getattr(settings, 'NORTH_FIXTURES_TPL',
                    migrations.fixtures_default_tpl)
            .format(init_version))
        self.run_script(fixtures_path)

    def run_script(self, path):
        with io.open(path, 'r', encoding='utf8') as f:
            script = Script(f)
            script.run(connection)



# -*- coding: utf-8 -*-
from django.contrib.staticfiles.management.commands.runserver import \
    Command as RunserverCommand

from django_north.management import migrations


class Command(RunserverCommand):
    help = ("Starts a lightweight Web server for development and also "
            "serves static files.")

    def check_migrations(self):
        try:
            migration_plan = migrations.build_migration_plan()
        except migrations.DBException as e:
            self.stdout.write(self.style.NOTICE("\n{}\n".format(e)))
            return

        if migration_plan is None:
            self.stdout.write(self.style.NOTICE("\nSchema not inited.\n"))
            return

        has_migrations = any(
            [
                any([not applied for mig, applied, path in plan['plan']])
                for plan in migration_plan['plans']
            ]
        )
        if has_migrations:
            self.stdout.write(self.style.NOTICE(
                "\nYou have unapplied migrations; your app may not work "
                "properly until they are applied."
            ))
            self.stdout.write(self.style.NOTICE(
                "Run 'python manage.py migrate' to apply them.\n"
            ))



from django.contrib.auth.management import _get_all_permissions
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db import DEFAULT_DB_ALIAS


def get_all_contenttypes_for_app_config(app_config, using=DEFAULT_DB_ALIAS):
    """
    Return all contenttypes tht already exists for a given app_config
    """
    ctypes = list()
    for klass in app_config.get_models():
        opts = ContentType.objects._get_opts(klass, True)
        try:
            ctype = ContentType.objects.db_manager(using).get(
                app_label=opts.app_label, model=opts.model_name)
            ctypes.append((ctype, klass))
        except ContentType.DoesNotExist:
            continue
    return ctypes


def get_searched_permissions(ctypes):
    """
    Return all permissions that should exist for existing contenttypes
    """
    # This will hold the permissions we're looking for as
    # (content_type, (codename, name))
    searched_perms = list()
    for ctype, klass in ctypes:
        for perm in _get_all_permissions(klass._meta, ctype):
            searched_perms.append((ctype, perm))

    return searched_perms


def get_all_permissions(ctypes, using=DEFAULT_DB_ALIAS):
    """
    Return all existing permissions for existing contenttypes
    """
    # Find all the Permissions that have a content_type for a model we're
    # looking for.  We don't need to check for codenames since we already have
    # a list of the ones we're going to create.
    return set(Permission.objects.using(using).filter(
        content_type__in=ctypes,
    ).values_list(
        "content_type", "codename"
    ))


def get_missing_permissions_for_app_config(app_config):
    """
    Return missing permissions for a given app config
    """
    # get existing contenttypes
    ctypes = get_all_contenttypes_for_app_config(app_config)
    if not ctypes:
        return []

    # get all permissions that should exist
    searched_perms = get_searched_permissions(ctypes)
    if not searched_perms:
        return []

    # get existing permissions
    all_perms = get_all_permissions([ctype for ctype, klass in ctypes])

    # build missing permissions
    perms = [
        Permission(codename=codename, name=name, content_type=ct)
        for ct, (codename, name) in searched_perms
        if (ct.pk, codename) not in all_perms
    ]

    return perms
